// Microcode Assembler.cpp : main project file.

#include "stdafx.h"
#include "MainForm.h"

using namespace MicrocodeAssembler;
using namespace System::IO;
using namespace System::Collections::Generic;

// // aluleft(2) | aluright(2) | load(3) | dralu  ||  aluop(4) | alumode | alucarry | enablecc | ld_shift_cc  ||  incmar | incpc | cntsp | spdir | ldop | adrsel(2) | nc

/*************************************************************************
/* Preset an open file dialog when the application starts.
/*
/*************************************************************************/
void MainForm::MainForm_Load(System::Object^  sender, System::EventArgs^  e) 
{
	if (s_bFirstLoad)
	{
		s_bFirstLoad = false;
		
		OpenFileDialog^ dialog = gcnew OpenFileDialog;

		//dialog->InitialDirectory = "c:\\";
		dialog->Filter = "microcode source files (*.usrc)|*.usrc|All files (*.*)|*.*";
		dialog->FilterIndex = 2;
		dialog->RestoreDirectory = true;

		if ( dialog->ShowDialog() == System::Windows::Forms::DialogResult::OK )
		{
			ParseMicrocode(dialog->FileName);
		}

		Close();
	}
}

/*************************************************************************
/* Initialize the microcode lines to "uninitialized" state.
/*
/*************************************************************************/
void MainForm::InitData(array<MicrocodeLine>^ romData)
{
	for (int i=0; i<kROMSize; i++)
	{
		romData[i].wasSet = false;
		// high byte: left = A, right = T, load A, drive ALU
		// mid byte: ALUfunc = A, retain existing CC's
		// low byte: nothing increments, don't ldop, ADRSEL=PC
		romData[i].data = 0x00FAFCFF;
	}
}

/*************************************************************************
/* Open the input file and parse each line, one at a time.
/*
/*************************************************************************/
void MainForm::ParseMicrocode(String^ filename)
{
	array<MicrocodeLine>^ romData = gcnew array<MicrocodeLine>(kROMSize);
	int nextAddress = 0;

	InitData(romData);

	try
	{
		StreamReader^ sr = gcnew StreamReader(filename);

		try
		{
			String^ line;
		 
			// Read and display lines from the file until the end of 
			// the file is reached.
			while ( (line = sr->ReadLine()) != nullptr && nextAddress >= 0 )
			{
				static bool parsedHalt = false;
				int prevAddress = nextAddress;
				nextAddress = ParseLine( line, nextAddress, romData );
				if (!parsedHalt && nextAddress < prevAddress)
				{
					SetAllInstructionsToHalt( romData );
					parsedHalt = true;
				}
			}

			if (nextAddress >= 0)
			{
				SaveData(filename, romData);
				MessageBox::Show( "Microcode data saved to ROM files.", "Success" );
			}
		}
		finally
		{
			if ( sr )
			delete (IDisposable^)sr;		
		}
	}
	catch ( Exception^ ) 
	{
		MessageBox::Show( "An error occurred while reading the input file.", "File Error" );
	}
}

void MainForm::SetAllInstructionsToHalt(array<MicrocodeLine>^ romData)
{
	for (int op = 0; op < 0xFF; op++)
	{
		for (int index = 0; index < 512; index++)
		{
			romData[op*512+index].data = romData[0xFF*512+index].data;
		}
	}
}

/*************************************************************************
/* Save the assembled data into 3 ROM files, in the same directory as the
/* input file.
/*
/*************************************************************************/
void MainForm::SaveData( String^ filename, array<MicrocodeLine>^ romData )
{
	// TODO: verify/sanity check
	// things to check for:
	// - some lines were not set, when higher ones were set
	// - instruction doesn't load next op on some paths
	// - can't read and write memory at the same time
	// - can't write PCLO, PCHI directly to memory

	String^ path = filename->Substring(0, filename->LastIndexOf("\\"));

	// write 3 ROM files
	for (int rom=0; rom<3; rom++)
	{
		try
		{
			StreamWriter^ sw = gcnew StreamWriter(path + String::Format("\\microcode_rom_{0}.rom", rom));

			try
			{
				int bytes=0;
				for (int address=0; address<kROMSize; address++)
				{
					String^ line = String::Format("{0:X6}", romData[address].data);
					sw->Write(String::Format("{0} ", line->Substring(2*(2-rom), 2)));
					
					bytes++;
					if (bytes == 32)
					{
						sw->WriteLine("");
						bytes = 0;
					}
				}
			}
			finally
			{
				if ( sw )
				delete (IDisposable^)sw;
			}
		}
		catch ( Exception^ ) 
		{
			MessageBox::Show( "An error occurred while saving the output files.", "File Error" );
		}
	}

	// write 3 HEX files
	for (int rom=0; rom<3; rom++)
	{
		try
		{
			StreamWriter^ sw = gcnew StreamWriter(path + String::Format("\\microcode_rom_{0}.hex", rom));

			try
			{		
				unsigned int addressHigh = 0;
				sw->WriteLine(":020000040000FA");
				for (int address=0; address<kROMSize; address+=16)
				{
					unsigned int newAddressHigh = ((address >> 16) & 0xFFFF);
					if (newAddressHigh != addressHigh)
					{
						unsigned int adrChecksum = 2 + 4 + newAddressHigh;
						adrChecksum = adrChecksum & 0xFF;
						adrChecksum = 0x100 - adrChecksum;
						adrChecksum = adrChecksum & 0xFF;
						sw->WriteLine(String::Format(":02000004{0:X4}{1:X2}", newAddressHigh, adrChecksum));	
						addressHigh = newAddressHigh;
					}

					sw->Write(String::Format(":10{0:X4}00", address & 0xFFFF));
					
					unsigned int checksum = 16 + (address & 0xFF) + ((address >> 8) & 0xFF);
					for (int offset=0; offset<16; offset++)
					{
						String^ line = String::Format("{0:X6}", romData[address+offset].data);
						String^ byte = String::Format("{0}", line->Substring(2*(2-rom), 2));
						sw->Write(byte);

						checksum += System::Convert::ToUInt32(byte, 16);
					}

					checksum = checksum & 0xFF;
					checksum = 0x100 - checksum;
					checksum = checksum & 0xFF;

					sw->WriteLine(String::Format("{0:X2}", checksum));
				}

				sw->WriteLine(":00000001FF");
			}
			finally
			{
				if ( sw )
				delete (IDisposable^)sw;
			}
		}
		catch ( Exception^ ) 
		{
			MessageBox::Show( "An error occurred while saving the output files.", "File Error" );
		}
	}
}

/*************************************************************************
/* Parse a single line of microcode, update the ROM data, and return the
/* next address to use.
/*
/* The line is expected to be something in the form:
/* - blank (trailing comments are OK)
/* - opcode NN (hex)
/* - C: A <- B + 2, OP <- MEM(PC), PC++
/*   where C is one of 0, 1, *
/*************************************************************************/
unsigned int MainForm::ParseLine( String^ line, unsigned int nextAddress, array<MicrocodeLine>^ romData )
{
	static bool newOp = true;
	static int phase = 0;
	static int lineNum = 0;
	lineNum++;
		
	// strip trailing comments
	int commentBegin = line->IndexOf("//");
	if (commentBegin != -1)
	{
		line = line->Substring(0, commentBegin);
	}
	line = line->Trim();

	// check for blank line
	if (line->Length == 0)
	{
		return nextAddress;
	}

	// check for "opcode:"
	if (line->StartsWith(kOpcodeStr, StringComparison::OrdinalIgnoreCase))
	{
		line = line->Substring(kOpcodeStr->Length);
		line = line->Trim();

		// get new ROM address
		int opcode = System::Convert::ToUInt32(line, 16);
		nextAddress = opcode * 512;

		// make sure it wasn't already used
		if (romData[nextAddress].wasSet)
		{
			MessageBox::Show(String::Format("Line {1}: redefinition of opcode {0:X2}", opcode, lineNum), "Parsing Error");
			nextAddress = -1;
		}

		newOp = true;
		phase = 0;
		return nextAddress;
	}

	// check for too many instructions in an opcode
	//if (!newOp && (nextAddress & 0xFF) == 0)
	if (phase >= 16)
	{
		MessageBox::Show(String::Format("Line {0}: opcode has more than 16 phases", lineNum), "Parsing Error");
		return -1;
	}

	newOp = false;
	
	// check for "c=0:" and parse the condition code
	int firstColon = line->IndexOf(":");
	if (firstColon == -1)
	{
		MessageBox::Show(String::Format("Line {1}: missing condition code on line \"{0}\"", line, lineNum), "Parsing Error");
		return -1;
	}

	bool result;
	if (line->Substring(0,2)->Equals("*:"))
	{
		// skip past the condition code prefix
		line = line->Substring(firstColon+1)->Trim();
		result = ParseLineCommands(lineNum, nextAddress, 0, 32, line, romData);
	}
	else
	{
		// split out the individual conditions
		array<wchar_t>^ separators = gcnew array<wchar_t>(1);
		separators[0] = L',';
		array<String^>^ conditions = line->Substring(0,firstColon)->Split(separators);
	
		List<unsigned int>^ conditionsData = gcnew List<unsigned int>();

		for each (String^ conditionString in conditions)
		{
			String^ condition = conditionString->Substring(0,1);
			String^ eq = conditionString->Substring(1,1);
			int cc = 0;
			try 
			{
				cc = System::Convert::ToUInt32(conditionString->Substring(2,1));
			}
			catch (Exception^)
			{
				cc = -1;
			}

			unsigned int bitpos = GetCondtionBit(condition);

			if ((cc != 0 && cc != 1) || !eq->Equals("=") || (bitpos < 0))
			{
				MessageBox::Show(String::Format("Line {1}: invalid condition code on line \"{0}\"", line, lineNum), "Parsing Error");
				return -1;
			}	

			conditionsData->Add((bitpos << 1) | cc);
		}

		// skip past the condition code prefix
		line = line->Substring(firstColon+1)->Trim();

		for (unsigned int count=0; count<32; count++)
		{
			bool conditionsMatch = true;
			for each (unsigned int data in conditionsData)
			{
				if (((1 << (data >> 1)) & count) != ((data & 1) << (data >> 1)))
				{
					conditionsMatch = false;
					break;
				}
			}

			if (conditionsMatch)
			{
				result = ParseLineCommands(lineNum, nextAddress, count, 1, line, romData);
				if (!result)
				{
					break;
				}
			}
		}
	}

	if (!result)
	{
		return -1;
	}

	// if all condition codes have been set, increment address
	bool allSet = true;
	for (int offset=0; offset<32; offset++)
	{
		if (!romData[nextAddress+offset].wasSet)
		{
			allSet = false;
			break;		
		}
	}

	if (allSet)
	{
		nextAddress += 32;
		phase++;
	}

	return nextAddress;
}

/*************************************************************************
/* Return the bit position of the desired condition code.
/*
/*************************************************************************/
unsigned int MainForm::GetCondtionBit(String^ condition)
{
	if (condition->Equals("a", StringComparison::OrdinalIgnoreCase))
	{
		return 0;
	}
	else if (condition->Equals("c", StringComparison::OrdinalIgnoreCase))
	{
		return 1;
	}
	else if (condition->Equals("n", StringComparison::OrdinalIgnoreCase))
	{
		return 2;
	}
	else if (condition->Equals("z", StringComparison::OrdinalIgnoreCase))
	{
		return 3;
	}
	else if (condition->Equals("v", StringComparison::OrdinalIgnoreCase))
	{
		return 4;
	}
	else 
	{
		return -1;
	}
}

/*************************************************************************
/* Check to make sure the instruction doesn't make conflicting use of
/* memory sources, such as MEM(SP) <- ARLO.
/*
/*************************************************************************/
bool MainForm::CheckConflictingAddressSources(int lineNum, unsigned int nextAddress, int cc, array<MicrocodeLine>^ romData, unsigned int sourceAddressType)
{
	unsigned int destinationAddressType = GetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit);

	if (sourceAddressType != destinationAddressType &&
		destinationAddressType != kAddressSelPC)
	{
		MessageBox::Show(String::Format("Line {0}: conflicting use of address bus sources", lineNum), "Parsing Error");
		return false;
	}

	return true;
}

/*************************************************************************
/* Parse the command portion of a line, setting microcode bits as required.
/*
/*************************************************************************/
bool MainForm::ParseLineCommands(int lineNum, unsigned int nextAddress, int cc, int numLines, String^ line, array<MicrocodeLine>^ romData)
{
	// split out the individual commands
	array<wchar_t>^ separators = gcnew array<wchar_t>(1);
	separators[0] = L';';
	array<String^>^ commands = line->Split(separators);

	for (int i=0; i<numLines; i++, cc++)
	{
		if (romData[nextAddress+cc].wasSet)
		{
			MessageBox::Show(String::Format("Line {1}: redefinition of phase on line \"{0}\"", line, lineNum), "Parsing Error");
			return false;
		}

		romData[nextAddress+cc].wasSet = true;

		// initial value
		romData[nextAddress+cc].data = 0x00FAFCFF;

		// check for some mutually exclusive commands
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kIncSPStr, kEnableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kDecSPStr, kDisableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kIncPCStr, kIncARStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kIncPCStr, kEnableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kLoadOpStr, kEnableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kLoadOpStr, kDisableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kIncPCStr, kDisableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kIncARStr, kEnableInterruptStr))
		{
			return false;
		}
		if (!CheckMutuallyExclusiveCommands(lineNum, commands, kIncARStr, kDisableInterruptStr))
		{
			return false;
		}

		for each (String^ command in commands)
		{
			command = command->Trim();
			
			if (command->Length == 0)
			{
				continue;
			}
			
			// check for sp++
			if (command->Equals(kIncSPStr, StringComparison::OrdinalIgnoreCase))
			{
				ClearBit(romData[nextAddress+cc].data, kCountSPBit);
			}
			// check for sp--
			else if (command->Equals(kDecSPStr, StringComparison::OrdinalIgnoreCase))
			{
				ClearBit(romData[nextAddress+cc].data, kCountSPBit);
				ClearBit(romData[nextAddress+cc].data, kSPDirIEDataBit);
			}
			// The next 4 are all mutually exclusive
			// check for pc++
			else if (command->Equals(kIncPCStr, StringComparison::OrdinalIgnoreCase))
			{
				ClearBit(romData[nextAddress+cc].data, kIncPCBit);
			}
			// check for ar++
			else if (command->Equals(kIncARStr, StringComparison::OrdinalIgnoreCase))
			{
				ClearBit(romData[nextAddress+cc].data, kIncARBit);
			}
			// check for interrupt enable
			else if (command->Equals(kEnableInterruptStr, StringComparison::OrdinalIgnoreCase))
			{
				ClearBit(romData[nextAddress+cc].data, kIncPCBit);
				ClearBit(romData[nextAddress+cc].data, kIncARBit);
				ClearBit(romData[nextAddress+cc].data, kSPDirIEDataBit);
			}
			// check for interrupt disable
			else if (command->Equals(kDisableInterruptStr, StringComparison::OrdinalIgnoreCase))
			{
				ClearBit(romData[nextAddress+cc].data, kIncPCBit);
				ClearBit(romData[nextAddress+cc].data, kIncARBit);
			}
			// check for nextop
			else if (command->Equals(kLoadOpStr, StringComparison::OrdinalIgnoreCase))
			{
				// translates to op <- mem(pc), pc++
				ClearBit(romData[nextAddress+cc].data, kLoadOpBit);
				ClearBit(romData[nextAddress+cc].data, kIncPCBit);
			}
			// check for latchNZ
			else if (command->Equals(kLatchNZStr, StringComparison::OrdinalIgnoreCase))
			{
				SetBit(romData[nextAddress+cc].data, kShifterS0Bit);
			}
			// check for latchAllCC
			else if (command->Equals(kLatchAllCCStr, StringComparison::OrdinalIgnoreCase))
			{
				unsigned int right = GetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit);
				unsigned int dest = GetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit);

				if (kLoadDestA!=dest || !GetBit(romData[nextAddress+cc].data, kLowDestBit) || kRightOpT != right)
				{
					MessageBox::Show(String::Format("Line {1}: can't latchAllCC unless dest=A and right=T \"{0}\"", command, lineNum), "Parsing Error");
					return false;
				}

				SetBit(romData[nextAddress+cc].data, kShifterS0Bit);
				SetBit(romData[nextAddress+cc].data, kShifterS1Bit);
			}
			// check for latchNZC
			else if (command->Equals(kLatchNZCStr, StringComparison::OrdinalIgnoreCase))
			{
				unsigned int right = GetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit);
				unsigned int dest = GetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit);

				if (kLoadDestA==dest && GetBit(romData[nextAddress+cc].data, kLowDestBit) && kRightOpT == right)
				{
					MessageBox::Show(String::Format("Line {1}: can't latchNZC when dest=A and right=T \"{0}\"", command, lineNum), "Parsing Error");
					return false;
				}

				SetBit(romData[nextAddress+cc].data, kShifterS0Bit);
				SetBit(romData[nextAddress+cc].data, kShifterS1Bit);
			}
			// check for shiftcc
			else if (command->Equals(kShiftCCStr, StringComparison::OrdinalIgnoreCase))
			{
				SetBit(romData[nextAddress+cc].data, kShifterS1Bit);			
			}
			// check for assignment
			else if (command->IndexOf(kAssignStr, StringComparison::OrdinalIgnoreCase) != -1)
			{
				// split to src and dest
				array<wchar_t>^ separators = gcnew array<wchar_t>(1);
				separators[0] = L'<';
				array<String^>^ ops = command->Split(separators);
				if (ops->Length != 2)
				{
					MessageBox::Show(String::Format("Line {1}: malformed assignment \"{0}\"", command, lineNum), "Parsing Error");
					return false;
				}

				// get the source(s) and dest
				String^ destStr = ops[0]->Trim();
				String^ sourcesStr = ops[1]->Substring(1)->Trim();

				if (!ParseDestination(lineNum, nextAddress, cc, destStr, romData))
				{
					return false;
				}

				// SRC
				// check for generic ALU function: ALU(L,R,OP,/C,M)
				if (sourcesStr->IndexOf(kALUStr, StringComparison::OrdinalIgnoreCase) != -1)
				{
					String^ paramsStr = sourcesStr->Substring(sourcesStr->IndexOf(kALUStr)+5)->Trim();
					
					// strip the trailing ")"
					int parenIndex = paramsStr->IndexOf(kParenStr, StringComparison::OrdinalIgnoreCase);
					if (parenIndex == -1)
					{
						MessageBox::Show(String::Format("Line {1}: malformed ALU operation \"{0}\"", sourcesStr, lineNum), "Parsing Error");
						return false;
					}
					paramsStr = paramsStr->Substring(0, parenIndex)->Trim();

					// split out the individual params
					array<wchar_t>^ separators = gcnew array<wchar_t>(1);
					separators[0] = L',';
					array<String^>^ params = paramsStr->Split(separators);

					if (params->Length != 5)
					{
						MessageBox::Show(String::Format("Line {1}: malformed ALU operation \"{0}\"", sourcesStr, lineNum), "Parsing Error");
						return false;
					}
					
					// left operand
					if (!ParseLeftOperand(nextAddress, cc, params[0], romData))
					{
						MessageBox::Show(String::Format("Line {1}: unrecognized left operand for ALU function \"{0}\"", params[0], lineNum), "Parsing Error");
						return false;
					}

					// right operand
					if (!ParseRightOperand(lineNum, nextAddress, cc, params[1], romData))
					{
						MessageBox::Show(String::Format("Line {1}: unrecognized right operand for ALU function \"{0}\"", params[1], lineNum), "Parsing Error");
						return false;
					}

					// function
					unsigned int func = System::Convert::ToUInt32(params[2], 16); // try hex first
					if (func > 15)
					{
						// try decimal if hex parsing result was too large. E.g. "12"
						func = System::Convert::ToUInt32(params[2]);
					}
					if (func > 15 || func < 0)
					{
						MessageBox::Show(String::Format("Line {1}: unrecognized function for ALU \"{0}\"", params[2], lineNum), "Parsing Error");
						return false;
					}

					// carry
					unsigned int cbar = System::Convert::ToUInt32(params[3]);
					if (cbar != 0 && cbar != 1)
					{
						MessageBox::Show(String::Format("Line {1}: unrecognized carry for ALU function \"{0}\"", params[3], lineNum), "Parsing Error");
						return false;
					}

					// mode
					unsigned int m = System::Convert::ToUInt32(params[4]);
					if (m != 0 && m != 1)
					{
						MessageBox::Show(String::Format("Line {1}: unrecognized mode for ALU function \"{0}\"", params[4], lineNum), "Parsing Error");
						return false;
					}

					unsigned int specialALU = (func << 2) | (m << 1) | cbar;
					SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, specialALU);
				}
				// check for special subtraction form
				else if (sourcesStr->IndexOf(kMinusStr) != -1)
				{
					if (!ParseAddSubtractSources(lineNum, nextAddress, cc, sourcesStr, kMinusStr, kALUFuncAminus1, kALUFuncAminusB, kALUFuncAminusBminus1, romData))
					{
						return false;
					}
				}
				// check for special addition form
				else if (sourcesStr->IndexOf(kPlusStr) != -1)
				{
					if (!ParseAddSubtractSources(lineNum, nextAddress, cc, sourcesStr, kPlusStr, kALUFuncAplus1, kALUFuncAplusB, kALUFuncAplusBplus1, romData))
					{
						return false;
					}
				}
				// assume a simple register move
				else 
				{
					String^ srcStr = sourcesStr;

					if (ParseLeftOperand(nextAddress, cc, srcStr, romData))
					{
						SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, kALUFuncA);
					}
					else if (ParseRightOperand(lineNum, nextAddress, cc, srcStr, romData))
					{
						SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, kALUFuncB);
					}
					// ? <- MEM(PC)
					else if (srcStr->Equals(kRegMEMPC, StringComparison::OrdinalIgnoreCase))
					{
						SetBit(romData[nextAddress+cc].data, kDriveALUBit);
					}
					// ? <- MEM(AR)
					else if (srcStr->Equals(kRegMEMAR, StringComparison::OrdinalIgnoreCase))
					{
						if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelAR))
						{
							SetBit(romData[nextAddress+cc].data, kDriveALUBit);
							SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelAR);
						}
					}
					// ? <- MEM(SP)
					else if (srcStr->Equals(kRegMEMSP, StringComparison::OrdinalIgnoreCase))
					{
						if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelSP))
						{
							SetBit(romData[nextAddress+cc].data, kDriveALUBit);
							SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelSP);
						}
					}
					else
					{
						MessageBox::Show(String::Format("Line {1}: unrecognized source \"{0}\"", srcStr, lineNum), "Parsing Error");
						return false;
					}
				}
			}

			else 
			{
				MessageBox::Show(String::Format("Line {1}: unrecognized command \"{0}\"", command, lineNum), "Parsing Error");
				return false;
			}
		}
	}

	return true;
}

/*************************************************************************
/* Check if two mutually-incompatible commands were used together. If so,
/* show an error message.
/*
/* return - True if usage is valid, false if mutually-incompatible commands 
/* were used together.
/*
/*************************************************************************/
bool MainForm::CheckMutuallyExclusiveCommands(int lineNum, array<String^>^ commands, String^ command1, String^ command2)
{
	for each (String^ commandA in commands)
	{
		if (commandA->Trim()->Equals(command1, StringComparison::OrdinalIgnoreCase))
		{
			for each (String^ commandB in commands)
			{
				if (commandB->Trim()->Equals(command2, StringComparison::OrdinalIgnoreCase))
				{
					MessageBox::Show(String::Format("Line {0}: can't combine {1} and {2}", lineNum, command1, command2), "Parsing Error");
					return false;
				}
			}
			break;
		}
	}
		
	return true;
}
/*************************************************************************
/* Check the input string for a valid destination name, and set load bits
/* as needed.
/*
/*************************************************************************/
bool MainForm::ParseDestination(int lineNum, unsigned int nextAddress, int cc, String^ destStr, array<MicrocodeLine>^ romData)
{
	// A
	if (destStr->Equals(kRegA, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestA);
	}
	// X
	else if (destStr->Equals(kRegX, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestX);
	}
	// Y
	else if (destStr->Equals(kRegY, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestY);
	}
	// T
	else if (destStr->Equals(kRegT, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestT);
	}
	// PCLO
	else if (destStr->Equals(kRegPCLO, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestPCLO);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// PCHI
	else if (destStr->Equals(kRegPCHI, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestPCHI);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// PCBANK
	else if (destStr->Equals(kRegPCBANK, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestPCBANK);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// ARLO
	else if (destStr->Equals(kRegARLO, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestARLO);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// ARHI
	else if (destStr->Equals(kRegARHI, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestARHI);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// ARBANK
	else if (destStr->Equals(kRegARBANK, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestARBANK);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// SPLO
	else if (destStr->Equals(kRegSPLO, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestSPLO);
	}
	// SPHI
	else if (destStr->Equals(kRegSPHI, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestSPHI);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// SPBANK
	else if (destStr->Equals(kRegSPBANK, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestSPBANK);
		ClearBit(romData[nextAddress+cc].data, kLowDestBit);
	}
	// MEM(PC)
	else if (destStr->Equals(kRegMEMPC, StringComparison::OrdinalIgnoreCase))
	{
		MessageBox::Show(String::Format("Line {1}: invalid destination \"{0}\"", destStr, lineNum), "Parsing Error");
		return false;
	}
	// MEM(AR)
	else if (destStr->Equals(kRegMEMAR, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestMem);
		SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelAR);
	}
	// MEM(SP)
	else if (destStr->Equals(kRegMEMSP, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLoadDestMask, kLoadDestBit, kLoadDestMem);
		SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelSP);
	}
	else
	{
		MessageBox::Show(String::Format("Line {1}: unrecognized destination \"{0}\"", destStr, lineNum), "Parsing Error");
		return false;
	}

	return true;
}

/*************************************************************************
/* Parse the sources for an add or subtract operation, setting ALU bits
/* as needed.
/*
/*************************************************************************/
bool MainForm::ParseAddSubtractSources(int lineNum, unsigned int nextAddress, int cc, String^ sourcesStr, String^ opString, unsigned int aluOpIncDec, unsigned int aluOpNormal, unsigned int aluOpThreeOperand, array<MicrocodeLine>^ romData)
{
	int firstIndex = sourcesStr->IndexOf(opString);
	String^ leftOp = sourcesStr->Substring(0, firstIndex)->Trim();
	String^ remainder = sourcesStr->Substring(firstIndex + 1)->Trim();

	if (!ParseLeftOperand(nextAddress, cc, leftOp, romData))
	{
		MessageBox::Show(String::Format("Line {1}: unrecognized primary add/sub operand \"{0}\"", leftOp, lineNum), "Parsing Error");
		return false;
	}

	// check for L + 1 form
	if (remainder->Equals(kOneStr, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, aluOpIncDec);
	}
	else
	{
		String^ rightOp;
		
		// check for L + R + 1 form
		bool threeTerms = false;
		if (remainder->IndexOf(opString, StringComparison::OrdinalIgnoreCase) != -1)
		{
			int secondIndex = remainder->IndexOf(opString);
			rightOp = remainder->Substring(0, secondIndex)->Trim();
			remainder = remainder->Substring(secondIndex + 1)->Trim();

			if (!remainder->Equals(kOneStr, StringComparison::OrdinalIgnoreCase))
			{
				MessageBox::Show(String::Format("Line {1}: unrecognized tertiary add/sub operand \"{0}\"", remainder, lineNum), "Parsing Error");
				return false;
			}
			else
			{
				SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, aluOpThreeOperand);
				threeTerms = true;
			}
		}
		else
		{
			rightOp = remainder;
			SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, aluOpNormal);
		}

		if (!ParseRightOperand(lineNum, nextAddress, cc, rightOp, romData))
		{
			// check for L + L or L + L + 1 special forms
			if (leftOp->Equals(rightOp) && aluOpIncDec == kALUFuncAplus1)
			{
				// set right operand to ARLO: it will be ignored anyway
				SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRLO);
				if (threeTerms)
				{
					SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, kALUFuncAplusAplus1);
				}
				else
				{
					SetField(romData[nextAddress+cc].data, kALUFuncMask, kALUFuncBit, kALUFuncAplusA);
				}	
			}
			else
			{
				MessageBox::Show(String::Format("Line {1}: unrecognized secondary add/sub operand \"{0}\"", rightOp, lineNum), "Parsing Error");
				return false;
			}
		}
	}

	return true;
}

/*************************************************************************
/* Check the input string for a valid left operand name, and set the 
/* ALU input bits if one is found.
/*
/*************************************************************************/
bool MainForm::ParseLeftOperand(unsigned int nextAddress, int cc, String^ leftOp, array<MicrocodeLine>^ romData)
{
	// parse L:
	// A
	if (leftOp->Equals(kRegA, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLeftOpMask, kLeftOpBit, kLeftOpA);
	}
	// X
	else if (leftOp->Equals(kRegX, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLeftOpMask, kLeftOpBit, kLeftOpX);
	}
	// Y
	else if (leftOp->Equals(kRegY, StringComparison::OrdinalIgnoreCase))
	{
		if (ShouldUseX7(nextAddress, cc, romData))
		{
			MessageBox::Show("Error: can't load address registers (except SPLO) from Y", "Parsing Error");
			return false;
		}
		else
		{
			SetField(romData[nextAddress+cc].data, kLeftOpMask, kLeftOpBit, kLeftOpY);
		}
	}
	// X7
	else if (leftOp->Equals(kRegX7, StringComparison::OrdinalIgnoreCase))
	{
		if (!ShouldUseX7(nextAddress, cc, romData))
		{
			MessageBox::Show("Error: can only load address registers (except SPLO) from X7", "Parsing Error");
			return false;
		}
		else
		{
			SetField(romData[nextAddress+cc].data, kLeftOpMask, kLeftOpBit, kLeftOpX7);
		}
	}
	// 0
	else if (leftOp->Equals(kReg0, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kLeftOpMask, kLeftOpBit, kLeftOp0);
	}
	else
	{
		return false;
	}

	return true;
}

/*************************************************************************
/* Check the input string for a valid right operand name, and set the 
/* ALU input bits if one is found.
/*
/*************************************************************************/
bool MainForm::ParseRightOperand(int lineNum, unsigned int nextAddress, int cc, String^ rightOp, array<MicrocodeLine>^ romData)
{
	// parse R:
	// T
	if (rightOp->Equals(kRegT, StringComparison::OrdinalIgnoreCase))
	{
		SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpT);
	}
	// PCLO
	else if (rightOp->Equals(kRegPCLO, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelPC))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRLO);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelPC);
		}
		else
		{
			return false;
		}
	}
	// PCHI
	else if (rightOp->Equals(kRegPCHI, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelPC))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRHI);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelPC);
		}
		else
		{
			return false;
		}
	}
	// PCBANK
	else if (rightOp->Equals(kRegPCBANK, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelPC))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRBANK);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelPC);
		}
		else
		{
			return false;
		}
	}
	// ARLO
	else if (rightOp->Equals(kRegARLO, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelAR))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRLO);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelAR);
		}
		else
		{
			return false;
		}
	}
	// ARHI
	else if (rightOp->Equals(kRegARHI, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelAR))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRHI);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelAR);
		}
		else
		{
			return false;
		}
	}
	// ARBANK
	else if (rightOp->Equals(kRegARBANK, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelAR))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRBANK);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelAR);
		}
		else
		{
			return false;
		}
	}
	// SPLO
	else if (rightOp->Equals(kRegSPLO, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelSP))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRLO);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelSP);
		}
		else
		{
			return false;
		}
	}
	// SPHI
	else if (rightOp->Equals(kRegSPHI, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelSP))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRHI);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelSP);
		}
		else
		{
			return false;
		}
	}
	// SPBANK
	else if (rightOp->Equals(kRegSPBANK, StringComparison::OrdinalIgnoreCase))
	{
		if (CheckConflictingAddressSources(lineNum, nextAddress, cc, romData, kAddressSelSP))
		{
			SetField(romData[nextAddress+cc].data, kRightOpMask, kRightOpBit, kRightOpADRBANK);
			SetField(romData[nextAddress+cc].data, kAddressSelMask, kAddressSelBit, kAddressSelSP);
		}
		else
		{
			return false;
		}
	}
	else
	{
		return false;
	}

	return true;
}

/*************************************************************************
/* Set or clear an individual bit.
/*
/*************************************************************************/
void MainForm::ClearBit(unsigned int% data, int bit)
{
	data &= ~(((unsigned int)1) << bit);
}
void MainForm::SetBit(unsigned int% data, int bit)
{
	data |= (((unsigned int)1) << bit);
}
int MainForm::GetBit(unsigned int% data, int bit)
{
	return ((data >> bit) & 1);
}
void MainForm::SetField(unsigned int% data, unsigned int mask, int bit, unsigned int value)
{
	data &= mask;
	data |= (value << bit);
}
unsigned int MainForm::GetField(unsigned int data, unsigned int mask, int bit)
{
	data &= (~mask);
	return data >> bit;
}

/*************************************************************************
/* Determine if X7 or Y will be used as the left operand, since they share
/* the same output enable bit in the microcode. X7 is used when an address
/* register is the destination, otherwise Y is used.
/*
/*************************************************************************/
bool MainForm::ShouldUseX7(unsigned int nextAddress, int cc, array<MicrocodeLine>^ romData)
{
	unsigned int loadDest0 = (romData[nextAddress+cc].data >> kLowDestBit) & 1;

	return (loadDest0 == 0);
}